// Перед запуском необходимо запустить кэширующую технологию Redis, используя команду ниже:
//redis-server /usr/local/etc/redis.conf
//Разработка программного интерфейса взаимодействия системы отображения метеорологических данных
// с базой данных измерительной информации
//Программный интерфейс на архитектуре Graphql
//Секция импорта библиотек и модулей
const createError = require('http-errors'); //подключение обработчика http-ошибок
const express = require('express'); //подключение фреймворка экспресс
const path = require('path'); //подключение модуля, определяющего путь файлов
const cookieParser = require('cookie-parser'); //подключение парсера cookie
const logger = require('morgan'); //подключение лоигурещего модуля
const redis = require('redis'); //подключение Redis
const moment = require('moment'); //подключение библиотеки для работы со временем
var express_graphql = require('express-graphql'); //Подключение библиотеки для совместной работы GraphQL и Express
var {buildSchema} = require('graphql'); //подключение модуля GraphQL
const request = require('request'); //подключение библиотеки для отправки асинхронных запросов
const fs = require('fs'); //Импорт модуля, позволяющего работать с файловой системой
//---Декларация всего необходимого для работы Redis
const port_redis = process.env.PORT || 6379; // Определение порта
const redis_client = redis.createClient(port_redis); //Создание подключения по выбранному порту
//---Конец
//При перезагрузке, повторном запуске очистить Redis
redis_client.flushdb(function (err, succeeded) {
    console.log(succeeded);
});
//---Блок подключения всего необходимого для фреймворка Экспресс---
var app = express();
app.set('views', path.join(__dirname, 'views'));
app.set('view engine', 'pug');
app.use(logger('dev'));
app.use(express.json());
app.use(express.urlencoded({extended: false}));
app.use(cookieParser());
app.use(express.static(path.join(__dirname, 'public')));
app.use(function (req, res, next) {
    next(createError(404));
});
// Обработчик ошибок
app.use(function (err, req, res, next) {
    res.locals.message = err.message;
    res.locals.error = req.app.get('env') === 'development' ? err : {};
    res.status(err.status || 500);
    res.render('error');
});
// ---Конец
/* API - BLOCK */

//Создаем схему API
//Определяем два вида запросов: откалиброванные и неоткалиброванные данные.
//Обязательно указать один из следующих типов времени:
//last5min(за последние 5 минут), lastweek(за последнюю неделю), log(за определенный интерфал), day(за определенный день)
//fdate и sdate указываются для определения временного интервала для log
//date указывается для определения времени для day
//uname и serial указываются в том случае, когда необходимо запросить данные с определенного прибора.
var schema = buildSchema(`
    type Query {
        calibr(time:String!,fdate:String,sdate:String,date:String,uname:String,serial:String):String!
        not_calibr(time:String!,fdate:String,sdate:String,date:String,uname:String,serial:String):String!
    }
`);




// Функция парсинга файла с калибровочными данными
// Возвращает имена приборов и их серийные номера, в случае, если для них есть калибровочные коэффициенты
function get_calibration_name_and_serial(file) {
    return new Promise((resolve, reject) => {
        fs.readFile(file, 'utf-8', function (error, data) { //reading file with json
            if (error) reject(error);
            let obj = JSON.parse(data);
            let array = Object.keys(obj);
            let respond_array = [];
            for (let i = 0; i < array.length; i++) {
                let respond_array_pair = [];
                let device_uname = obj[array[i]].uName;
                let device_serial = obj[array[i]].serial;
                respond_array_pair.push(device_uname);
                respond_array_pair.push(device_serial);
                respond_array.push(respond_array_pair);
            }
            resolve(respond_array);
        });
    });
}

//Функция поиска заданного прибора в списке калибровочных приборов, указав его uname и serial
// В случае ненахождения вернет -1, в противном случае вернет индекс из массива приборов, доступных для калибровки
function device_finder(uname, serial, device_list) {
    return new Promise((resolve, reject) => {
        let index;
        let request_uname_serial = [uname, serial];
        for (let i = 0; i < device_list.length; i++) {
            if (device_list[i][0] === request_uname_serial[0] && device_list[i][1] === request_uname_serial[1]) {
                index = i;
                resolve(index); //Вернуть индекс в массиве приборов, доступных для калибровки
            }
        }
        resolve(-1);
    });
}

// Функция чтения калибровочного файла
// Возвращает содержимое файла
function read_calibration_file(file) {
    return new Promise((resolve, reject) => {
        fs.readFile(file, 'utf-8', function (error, data) { //reading file with json
            if (error) reject(error);
            let obj = JSON.parse(data);
            resolve(obj); //Вернуть содержимое калибровочного файла
        });
    });
}


// Возвращает массив с названиями сенсоров для калибровки
function get_calibration_sensorname(file) {
    return new Promise((resolve, reject) => {
        fs.readFile(file, 'utf-8', function (error, data) { //reading file with json
            if (error) reject(error);
            let obj = JSON.parse(data);
            let array = Object.keys(obj);
            let sensor_array = [];
            let resp_array = [];
            for (let i = 0; i < obj.length; i++) {
                let sensors = obj[i].sensors;
                for (let j = 0; j < sensors.length; j++) {
                    sensor_array.push(sensors[j].sensor);
                }
                resp_array.push(sensor_array);
                sensor_array = [];
            }
            resolve(resp_array);
        });
    });
}


// С таймаутом 600000 делать запрос к базе, на случай если появятся новые калибровочные данные.
// Таким образом будет производиться учет регулярных обновлений калибровочных коэффициентов
var calibration_data = setInterval(function () {
    request({
        url: "http://188.35.161.31/core/deb.php?calibration",
        method: "GET", //Запрос типа GET
        timeout: 600000,
        followRedirect: true,
        maxRedirects: 10
    }, function (error, response, body) {
        if (!error && response.statusCode == 200) {
            console.log('Calibration Data has been loaded');
            fs.writeFile("CalibrationData.txt", response.body.toString(), function (error) {
                if (error) throw error;
            });

        } else {
            console.log('error' + response.statusCode);
        }
    });
}, 600000);


// Запись в кеш и лог-файл на сервере для данных за последние 5 минут
var requestLoop_five_min = setInterval(function () {
    //------Блок, отвечающий за формирование времени//
    let now = moment();
    let past = now.subtract(5, 'minutes');
    past = past.format("YYYY-MM-DD%20HH:mm:ss");
    let real_now = moment();
    real_now = real_now.format("YYYY-MM-DD%20HH:mm:ss");
    let base_requester = past + "&sdate=" + real_now;
//---Конец---//
    request({ //Делаем запрос к базе, получая данные за последние 5 минут
        url: "http://188.35.161.31/core/deb.php?fdate=" + base_requester,
        method: "GET", //Метод запроса - GET
        timeout: 5000, // выставление таймаута
        followRedirect: true,
        maxRedirects: 10  //максимальная попытка переподключений в случае обвала соединения
    }, function (error, response, body) {
        if (!error && response.statusCode == 200) {
            redis_client.set("Graghql#Not_calibr#LastLog", response.body.toString(), 'EX', 30); //Запомнить в редис
            fs.writeFile("LastLog.txt", response.body.toString(), function (error) { //записать в лог-файл
                if (error) throw error;
            });

        } else {
            console.log('error' + response.statusCode);
        }
    });
}, 30000);

// Запись в кеш и лог-файл на сервере для данных за последние 7 дней
var requestLoop_weekly = setInterval(function () {
    //Последние 7 дней
    let now_week = moment();
    let real_now_week = moment();
    real_now_week = real_now_week.format("YYYY-MM-DD%20HH:mm:ss");
    let past_seven_days = now_week.clone().subtract(7, 'days');
    past_seven_days = past_seven_days.format("YYYY-MM-DD%20HH:mm:ss");
    let base_requester_sevendays = past_seven_days + "&sdate=" + real_now_week;
//----
    request({
        url: "http://188.35.161.31/core/deb.php?fdate=" + base_requester_sevendays,
        method: "GET", //запрос типа GET
        timeout: 5000,
        followRedirect: true,
        maxRedirects: 10
    }, function (error, response, body) {
        if (!error && response.statusCode == 200) {
            redis_client.set("Graphql#Not_calibr#Week", response.body.toString(), 'EX', 30); //запись в редис
            fs.writeFile("WeekLog.txt", response.body.toString(), function (error) { //запись в лог-файл
                if (error) throw error;
            });

        }
    });
}, 30000);

//Функция обработки запросов на получение ВСЕХ НЕОТКАЛИБРОВАННЫХ ДАННЫХ
function req_not_calibr_all(url, redis_key) {
    return new Promise((resolve, reject) => {
        //Если в redis есть данные
        return redis_client.get(redis_key, (err, data) => {
            if (data) {
                resolve(data) //Вернуть из Redis
            } else {
                request(url, (error, response, body) => { //Иначе сделать запрос к базе
                    if (error) reject(error);
                    if (response.statusCode != 200) {
                        reject(response);
                    }
                    redis_client.set(redis_key, body, 'EX', 30); //Сохранить в redis
                    resolve(body);//Вернуть тело ответа
                });
            }
        })
    });
}

//Функция обработки запросов на получение ВСЕХ ПОСЛЕДНИХ ИЗМЕРЕНИЙ ДЛЯ ПРИБОРОВ
function req_not_calibr_lastmeasurement_all(url, redis_key) {
    return new Promise((resolve, reject) => {
        //Если в redis есть данные
        return redis_client.get(redis_key, (err, data) => {
            if (data) {
                resolve(data) //Вернуть из Redis
            } else {
                request(url, (error, response, body) => { //Иначе сделать запрос к базе
                    if (error) reject(error);
                    if (response.statusCode != 200) {
                        reject(response);
                    }
                    const respond = body;
                    let respond_js = JSON.parse(respond);
                    let array = Object.keys(respond_js);
                    array = array.reverse();
                    let respond_array = {};
                    for (let i = 0; i < array.length; i++) {
                        if (respond_array[respond_js[array[i]].uName]===undefined) {
                            let key = respond_js[array[i]].uName;
                            let key_pair = array[i];
                            let value_pair = respond_js[key_pair];
                            respond_array[key] = value_pair;
                        }
                    }
                    redis_client.set(redis_key, respond_array, 'EX', 30); //Сохранить в redis
                    resolve(respond_array);//Вернуть тело ответа
                });
            }
        })
    });
}

// Функция обработки запросов на получение НЕОТКАЛИБРОВОННОГО ПОСЛЕДНЕГО ИЗМЕРЕНИЯ С КОНКРЕТНОГО ПРИБОРА
function req_not_calibr_lastmeasurement_device(url, redis_key, request_uname, request_serial) {
    return new Promise((resolve, reject) => {
        //Если данные есть в redis
        return redis_client.get(redis_key, (err, data) => {
            if (data) {
                resolve(data) //Отправить данные из redis
            } else { //Иначе
                request(url, (error, response, body) => { //Сделать запрос к базе
                    if (error) reject(error);
                    if (response.statusCode != 200) {
                        reject(response);
                    }
                    const respond = body;
                    let respond_js = JSON.parse(respond);
                    let array = Object.keys(respond_js);
                    array = array.reverse();
                    let respond_array = {};
                    for (let i = 0; i < array.length; i++) {
                        if ((respond_js[array[i]].uName === request_uname) && (respond_js[array[i]].serial === request_serial)) {
                            let key_pair = array[i];
                            let value_pair = respond_js[key_pair];
                            respond_array[respond_js[array[i]].uName] = value_pair;
                            break;
                        }
                    }
                    redis_client.set(redis_key, JSON.stringify(respond_array), 'EX', 30); //сохранить данные в редис
                    resolve(respond_array);
                });

            }

        })
    });
}

// Функция обработки запросов на получение НЕОТКАЛИБРОВАННЫХ ДАННЫХ С КОНКРЕТНОГО ПРИБОРА
function req_not_calibr_device(url, redis_key, request_uname, request_serial) {
    return new Promise((resolve, reject) => {
        //Если данные есть в redis
        return redis_client.get(redis_key, (err, data) => {
            if (data) {
                resolve(data) //Отправить данные из redis
            } else { //Иначе
                request(url, (error, response, body) => { //Сделать запрос к базе
                    if (error) reject(error);
                    if (response.statusCode != 200) {
                        reject(response);
                    }
                    let respond_js = JSON.parse(body);
                    let array = Object.keys(respond_js);
                    let respond_array = {};
                    //Отсортировать данные по прибору и выдать только нужные
                    for (let i = 0; i < array.length; i++) {
                        if ((respond_js[array[i]].uName === request_uname) && (respond_js[array[i]].serial === request_serial)) {
                            let key_pair = array[i];
                            let value_pair = respond_js[key_pair];
                            respond_array[key_pair] = value_pair;
                        }
                    }
                    redis_client.set(redis_key, JSON.stringify(respond_array), 'EX', 30); //сохранить данные в редис
                    resolve(JSON.stringify(respond_array)); //Отправить данные
                });

            }

        })
    });
}
//Функция обработки запросов на получение ВСЕХ ОТКАЛИБРОВАННЫХ данных
function req_calibr_all(url, redis_key) {
    return new Promise((resolve, reject) => {
        //Если данные есть в Redis
        return redis_client.get(redis_key, (err, data) => {
            if (data) {
                resolve(data) //Отправить данные из Redis
            } else { //Иначе
                request(url, (error, response, body) => { //Сделать запрос к БД
                    if (error) reject(error);
                    if (response.statusCode != 200) {
                        reject(response);
                    }
                    let respond = body;
                    get_calibration_sensorname('CalibrationData.txt').then(function (result) {
                        let respond_sensor_name = result;
                        read_calibration_file('CalibrationData.txt').then(function (result) {
                            let calibration_file_content = result;
                            let respond_js = JSON.parse(respond);
                            let array = Object.keys(respond_js);
                            get_calibration_name_and_serial('CalibrationData.txt').then(function (result) {
                                let available_devices = result;
                                for (let i = 0; i < array.length; i++) {
                                    //Если запись получена с прибора, для которого доступна калибровка, то вернуть индекс прибора в массиве avalable_devices, иначе вернуть -1
                                    let index;
                                    let request_uname_serial = [respond_js[array[i]].uName, respond_js[array[i]].serial];
                                    for (let i = 0; i < available_devices.length; i++) {
                                        if (available_devices[i][0] === request_uname_serial[0] && available_devices[i][1] === request_uname_serial[1]) {
                                            index = i;
                                            break;
                                        }
                                        index = -1;
                                    }
                                    //Если прибор был найден в списке приборов, для которых есть калибровочные коэффициенты
                                    if (index !== -1) {
                                        let date_array = respond_sensor_name[index];
                                        for (let k = 0; k < date_array.length; k++) { //Для каждого сенсора, имеющего коэффициент
                                            let sensor_name = date_array[k];
                                            let sensor_index = Object.keys(respond_js[array[i]].data).indexOf(sensor_name);
                                            let calibr_datetime;
                                            let measurement_datetime = moment(new Date(respond_js[array[i]].Date));
                                            for(let x = 0; x <calibration_file_content[index].sensors[k].calibr.length;x++)
                                            {
                                                calibr_datetime = JSON.stringify(calibration_file_content[index].sensors[k].calibr[x].datetime);
                                                calibr_datetime = moment(new Date(calibr_datetime));
                                                if (measurement_datetime.isAfter(calibr_datetime) === false){
                                                    let calibr_n = JSON.stringify(calibration_file_content[index].sensors[k].calibr[x].data.n);
                                                    if (calibr_n == 1) {
                                                        calibr_ai = calibration_file_content[index].sensors[k].calibr[x].data.ai[0];
                                                        if (calibr_ai !== undefined) {
                                                            let obj = respond_js[array[i]].data;
                                                            let arr = Object.entries(obj);
                                                            if (arr[sensor_index] !== undefined) {
                                                                arr[sensor_index][1] = (parseFloat(arr[sensor_index][1]) + calibr_ai).toFixed(2);
                                                            }
                                                            respond_js[array[i]].data = Object.fromEntries(arr);
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                redis_client.set(redis_key, JSON.stringify(respond_js), 'EX', 30); //сохранить данные в редис
                                resolve(JSON.stringify(respond_js)) ? JSON.stringify(respond_js) : {};//отправить клиенту ответ на его запрос
                            });
                        })
                    });

                });

            }

        })
    });
}
//Функция обработки запросов на получение ОТКАЛИБРОВАННЫХ ДАННЫХ С КОКНЕРТНОГО ПРИБОРА
function req_calibr_device(url, redis_key, request_uname, request_serial) {
    return new Promise((resolve, reject) => {
        //Если данные есть в redis
        return redis_client.get(redis_key, (err, data) => {
            if (data) {
                resolve(data) //Отправить данные из Redis
            } else {
                request(url, (error, response, body) => { //Иначе сделать запрос к БД
                    if (error) reject(error);
                    if (response.statusCode != 200) {
                        reject(response);
                    }
                    let respond = body;
                    get_calibration_sensorname('CalibrationData.txt').then(function (result) {
                        let respond_sensor_name = result;
                        read_calibration_file('CalibrationData.txt').then(function (result) {
                            let calibration_file_content = result;
                            let respond_js = JSON.parse(respond);
                            let array = Object.keys(respond_js);
                            get_calibration_name_and_serial('CalibrationData.txt').then(function (result) {
                                let available_devices = result;
                                //----Блок поиска запрашиваемого прибора в массиве приборов, доступных для калибровки
                                let index;
                                let request_uname_serial = [request_uname, request_serial];
                                for (let i = 0; i < available_devices.length; i++) {
                                    if (available_devices[i][0] === request_uname_serial[0] && available_devices[i][1] === request_uname_serial[1]) {
                                        index = i; //Если нашелся, i=index
                                        break;
                                    }
                                    index = -1; //Если не нашелся, то i=-1
                                }
                                //----Конец
                                //---Блок формирования только тех записей, которые были получены с запрашиваемого прибора
                                let respond_array = {};
                                for (let i = 0; i < array.length; i++) {
                                    if ((respond_js[array[i]].uName === request_uname) && (respond_js[array[i]].serial === request_serial)) {
                                        var key_pair = array[i];
                                        var value_pair = respond_js[key_pair];
                                        respond_array[key_pair] = value_pair;
                                    }
                                }
                                //---Конец
                                if (index !== -1) { //Если прибор был найдем в списке приборов, доступных для калибровки
                                    array = Object.keys(respond_array);
                                    for (let i = 0; i < array.length; i++) {
                                        let date_array = respond_sensor_name[index];
                                        for (let k = 0; k < date_array.length; k++) { //Для каждого датчика, для которого есть калибровочный коэффициент
                                            let sensor_name = date_array[k];
                                            let sensor_index = Object.keys(respond_array[array[i]].data).indexOf(sensor_name);
                                            let calibr_datetime;
                                            let measurement_datetime = moment(new Date(respond_js[array[i]].Date));
                                            for(let x = 0; x <calibration_file_content[index].sensors[k].calibr.length;x++)
                                            {
                                                calibr_datetime = JSON.stringify(calibration_file_content[index].sensors[k].calibr[x].datetime);
                                                calibr_datetime = moment(new Date(calibr_datetime));
                                                if (measurement_datetime.isAfter(calibr_datetime) === false){
                                                    let calibr_n = JSON.stringify(calibration_file_content[index].sensors[k].calibr[x].data.n);
                                                    if (calibr_n == 1) {
                                                        calibr_ai = calibration_file_content[index].sensors[k].calibr[x].data.ai[0];
                                                        if (calibr_ai !== undefined) {
                                                            let obj = respond_js[array[i]].data;
                                                            let arr = Object.entries(obj);
                                                            if (arr[sensor_index] !== undefined) {
                                                                arr[sensor_index][1] = (parseFloat(arr[sensor_index][1]) + calibr_ai).toFixed(2);
                                                            }
                                                            respond_js[array[i]].data = Object.fromEntries(arr);
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                redis_client.set(redis_key, JSON.stringify(respond_array), 'EX', 30); //сохранить данные в редис
                                resolve(JSON.stringify(respond_array)) ? JSON.stringify(respond_array) : {};//отправить клиенту ответ на его запрос
                            });
                        })
                    });

                });

            }

        })
    });
}


//Функция обработки запросов на получение ОТКАЛИБРОВАННЫХ ДАННЫХ С КОКНЕРТНОГО ПРИБОРА
function req_calibr_lastmeasurement_device(url, redis_key, request_uname, request_serial) {
    return new Promise((resolve, reject) => {
        //Если данные есть в redis
        return redis_client.get(redis_key, (err, data) => {
            if (data) {
                resolve(data) //Отправить данные из Redis
            } else {
                request(url, (error, response, body) => { //Иначе сделать запрос к БД
                    if (error) reject(error);
                    if (response.statusCode != 200) {
                        reject(response);
                    }
                    let respond = body;
                    get_calibration_sensorname('CalibrationData.txt').then(function (result) {
                        let respond_sensor_name = result;
                        read_calibration_file('CalibrationData.txt').then(function (result) {
                            let calibration_file_content = result;
                            let respond_js = JSON.parse(respond);
                            let array = Object.keys(respond_js);
                            get_calibration_name_and_serial('CalibrationData.txt').then(function (result) {
                                let available_devices = result;
                                //----Блок поиска запрашиваемого прибора в массиве приборов, доступных для калибровки
                                let index;
                                let request_uname_serial = [request_uname, request_serial];
                                for (let i = 0; i < available_devices.length; i++) {
                                    if (available_devices[i][0] === request_uname_serial[0] && available_devices[i][1] === request_uname_serial[1]) {
                                        index = i; //Если нашелся, i=index
                                        break;
                                    }
                                    index = -1; //Если не нашелся, то i=-1
                                }
                                //----Конец
                                //---Блок формирования только тех записей, которые были получены с запрашиваемого прибора
                                array = array.reverse();
                                let respond_array = {};
                                for (let i = 0; i < array.length; i++) {
                                    if ((respond_js[array[i]].uName === request_uname) && (respond_js[array[i]].serial === request_serial)) {
                                        if (respond_array[respond_js[array[i]].uName]===undefined) {
                                            let key = respond_js[array[i]].uName;
                                            let key_pair = array[i];
                                            let value_pair = respond_js[key_pair];
                                            respond_array[key] = value_pair;
                                        }
                                    }
                                }
                                //---Конец
                                if (index !== -1) { //Если прибор был найдем в списке приборов, доступных для калибровки
                                    array = Object.keys(respond_array);
                                    for (let i = 0; i < array.length; i++) {
                                        let date_array = respond_sensor_name[index];
                                        for (let k = 0; k < date_array.length; k++) { //Для каждого датчика, для которого есть калибровочный коэффициент
                                            let sensor_name = date_array[k];
                                            let sensor_index = Object.keys(respond_array[array[i]].data).indexOf(sensor_name);
                                            let calibr_datetime;
                                            let measurement_datetime = moment(new Date(respond_js[array[i]].Date));
                                            for(let x = 0; x <calibration_file_content[index].sensors[k].calibr.length;x++)
                                            {
                                                calibr_datetime = JSON.stringify(calibration_file_content[index].sensors[k].calibr[x].datetime);
                                                calibr_datetime = moment(new Date(calibr_datetime));
                                                if (measurement_datetime.isAfter(calibr_datetime) === false){
                                                    let calibr_n = JSON.stringify(calibration_file_content[index].sensors[k].calibr[x].data.n);
                                                    if (calibr_n == 1) {
                                                        calibr_ai = calibration_file_content[index].sensors[k].calibr[x].data.ai[0];
                                                        if (calibr_ai !== undefined) {
                                                            let obj = respond_js[array[i]].data;
                                                            let arr = Object.entries(obj);
                                                            if (arr[sensor_index] !== undefined) {
                                                                arr[sensor_index][1] = (parseFloat(arr[sensor_index][1]) + calibr_ai).toFixed(2);
                                                            }
                                                            respond_js[array[i]].data = Object.fromEntries(arr);
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                redis_client.set(redis_key, JSON.stringify(respond_array), 'EX', 30); //сохранить данные в редис
                                resolve(JSON.stringify(respond_array)) ? JSON.stringify(respond_array) : {};//отправить клиенту ответ на его запрос
                            });
                        })
                    });

                });

            }

        })
    });
}




//Функция обработки запросов на получение ВСЕХ ПОСЛЕДНИХ ОТКАЛИБРОВАННЫХ ИЗМЕРЕНИЙ данных
function req_calibr_lastmeasurement_all(url, redis_key) {
    return new Promise((resolve, reject) => {
        //Если данные есть в Redis
        return redis_client.get(redis_key, (err, data) => {
            if (data) {
                resolve(data) //Отправить данные из Redis
            } else { //Иначе
                request(url, (error, response, body) => { //Сделать запрос к БД
                    if (error) reject(error);
                    if (response.statusCode != 200) {
                        reject(response);
                    }
                    let respond = body;
                    get_calibration_sensorname('CalibrationData.txt').then(function (result) {
                        let respond_sensor_name = result;
                        read_calibration_file('CalibrationData.txt').then(function (result) {
                            let calibration_file_content = result;
                            let respond_js = JSON.parse(respond);
                            let array = Object.keys(respond_js);
                            array = array.reverse();
                            let respond_array = {};
                            for (let i = 0; i < array.length; i++) {
                                if (respond_array[respond_js[array[i]].uName]===undefined) {
                                    let key = respond_js[array[i]].uName;
                                    let key_pair = array[i];
                                    let value_pair = respond_js[key_pair];
                                    respond_array[key] = value_pair;
                                }
                            }
                            array = Object.keys(respond_array);
                            respond_js = respond_array;
                            get_calibration_name_and_serial('CalibrationData.txt').then(function (result) {
                                let available_devices = result;
                                for (let i = 0; i < array.length; i++) {
                                    //Если запись получена с прибора, для которого доступна калибровка, то вернуть индекс прибора в массиве avalable_devices, иначе вернуть -1
                                    let index;
                                    let request_uname_serial = [respond_js[array[i]].uName, respond_js[array[i]].serial];
                                    for (let i = 0; i < available_devices.length; i++) {
                                        if (available_devices[i][0] === request_uname_serial[0] && available_devices[i][1] === request_uname_serial[1]) {
                                            index = i;
                                            break;
                                        }
                                        index = -1;
                                    }
                                    //Если прибор был найден в списке приборов, для которых есть калибровочные коэффициенты
                                    if (index !== -1) {
                                        let date_array = respond_sensor_name[index];
                                        for (let k = 0; k < date_array.length; k++) { //Для каждого сенсора, имеющего коэффициент
                                            let sensor_name = date_array[k];
                                            let sensor_index = Object.keys(respond_js[array[i]].data).indexOf(sensor_name);
                                            let calibr_datetime;
                                            let measurement_datetime = moment(new Date(respond_js[array[i]].Date));
                                            for(let x = 0; x <calibration_file_content[index].sensors[k].calibr.length;x++)
                                            {
                                                calibr_datetime = JSON.stringify(calibration_file_content[index].sensors[k].calibr[x].datetime);
                                                calibr_datetime = moment(new Date(calibr_datetime));
                                                if (measurement_datetime.isAfter(calibr_datetime) === false){
                                                    let calibr_n = JSON.stringify(calibration_file_content[index].sensors[k].calibr[x].data.n);
                                                    if (calibr_n == 1) {
                                                        calibr_ai = calibration_file_content[index].sensors[k].calibr[x].data.ai[0];
                                                        if (calibr_ai !== undefined) {
                                                            let obj = respond_js[array[i]].data;
                                                            let arr = Object.entries(obj);
                                                            if (arr[sensor_index] !== undefined) {
                                                                arr[sensor_index][1] = (parseFloat(arr[sensor_index][1]) + calibr_ai).toFixed(2);
                                                            }
                                                            respond_js[array[i]].data = Object.fromEntries(arr);
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                redis_client.set(redis_key, JSON.stringify(respond_js), 'EX', 30); //сохранить данные в редис
                                resolve(JSON.stringify(respond_js)) ? JSON.stringify(respond_js) : {};//отправить клиенту ответ на его запрос
                            });
                        })
                    });

                });

            }

        })
    });
}

// корневой узел API GraphQL
var root = {
    not_calibr: ({time, fdate, sdate, date, uname, serial}) => { //Неоткалиброванные данные
        if (time === 'last5min') { //Данные за последние 5 минут
            if (uname === undefined && serial === undefined) { //Если uname и serial не определены, то вернуть ВСЕ ИЗМЕРЕНИЯ
                //------Блок, формирующий время за последние 5 минут----//
                let now = moment();
                let past = now.subtract(5, 'minutes');
                past = past.format("YYYY-MM-DD%20HH:mm:ss");
                let real_now = moment();
                real_now = real_now.format("YYYY-MM-DD%20HH:mm:ss");
                //---Конец---//
                let base_requester = past + "&sdate=" + real_now;
                let url = "http://188.35.161.31/core/deb.php?fdate=" + base_requester; //Формируем URL для запроса к БД
                let redis_key = "Graphql" + "#" + "Not_calibr#LastLog"; //Формируем ключ, для хранения в Redis
                return req_not_calibr_all(url, redis_key);
            }
            if (uname !== undefined && serial !== undefined) { //Если uname и serial прибора были указаны
                let request_uname = uname;
                let request_serial = serial;
                //------Блок, отвечающий формирование времени ----//
                let now = moment();
                let past = now.subtract(5, 'minutes');
                past = past.format("YYYY-MM-DD%20HH:mm:ss");
                let real_now = moment();
                real_now = real_now.format("YYYY-MM-DD%20HH:mm:ss");
                //---Конец---//
                let base_requester = past + "&sdate=" + real_now;
                let url = "http://188.35.161.31/core/deb.php?fdate=" + base_requester; //Генерируем URL для запроса к БД
                let redis_key = "Graphql" + "#" + "Not_calibr" + "#" + "LastLog" + "#" + request_uname + "#" + request_serial; //Генерируем ключ Redis
                return req_not_calibr_device(url, redis_key, request_uname, request_serial);
            }
        }
        if (time === 'lastmin') { //Данные за последнюю минуту
            if (uname === undefined && serial === undefined) { //Если uname и serial не определены, то вернуть ВСЕ ИЗМЕРЕНИЯ
                //------Блок, формирующий время за последние 5 минут----//
                let now = moment();
                let past = now.subtract(1, 'minutes');
                past = past.format("YYYY-MM-DD%20HH:mm:ss");
                let real_now = moment();
                real_now = real_now.format("YYYY-MM-DD%20HH:mm:ss");
                //---Конец---//
                let base_requester = past + "&sdate=" + real_now;
                let url = "http://188.35.161.31/core/deb.php?fdate=" + base_requester; //Формируем URL для запроса к БД
                let redis_key = "Graphql" + "#" + "Not_calibr#LastMin"; //Формируем ключ, для хранения в Redis
                return req_not_calibr_all(url, redis_key);
            }
            if (uname !== undefined && serial !== undefined) { //Если uname и serial прибора были указаны
                let request_uname = uname;
                let request_serial = serial;
                //------Блок, отвечающий формирование времени ----//
                let now = moment();
                let past = now.subtract(1, 'minutes');
                past = past.format("YYYY-MM-DD%20HH:mm:ss");
                let real_now = moment();
                real_now = real_now.format("YYYY-MM-DD%20HH:mm:ss");
                //---Конец---//
                let base_requester = past + "&sdate=" + real_now;
                let url = "http://188.35.161.31/core/deb.php?fdate=" + base_requester; //Генерируем URL для запроса к БД
                let redis_key = "Graphql" + "#" + "Not_calibr" + "#" + "LastMin" + "#" + request_uname + "#" + request_serial; //Генерируем ключ Redis
                return req_not_calibr_device(url, redis_key, request_uname, request_serial);
            }
        }
        if (time === 'last_measurement') { //Данные за последнюю минуту
            if (uname === undefined && serial === undefined) { //Если uname и serial не определены, то вернуть ВСЕ ИЗМЕРЕНИЯ
                //------Блок, формирующий время за последние 5 минут----//
                let now = moment();
                let past = now.subtract(60, 'minutes');
                past = past.format("YYYY-MM-DD%20HH:mm:ss");
                let real_now = moment();
                real_now = real_now.format("YYYY-MM-DD%20HH:mm:ss");
                //---Конец---//
                let base_requester = past + "&sdate=" + real_now;
                let url = "http://188.35.161.31/core/deb.php?fdate=" + base_requester; //Формируем URL для запроса к БД
                let redis_key = "Graphql" + "#" + "Not_calibr#LastMeasurement"; //Формируем ключ, для хранения в Redis
                return req_not_calibr_lastmeasurement_all(url, redis_key);
            }
            if (uname !== undefined && serial !== undefined) { //Если uname и serial прибора были указаны
                let request_uname = uname;
                let request_serial = serial;
                //------Блок, отвечающий формирование времени ----//
                let now = moment();
                let past = now.subtract(1, 'minutes');
                past = past.format("YYYY-MM-DD%20HH:mm:ss");
                let real_now = moment();
                real_now = real_now.format("YYYY-MM-DD%20HH:mm:ss");
                //---Конец---//
                let base_requester = past + "&sdate=" + real_now;
                let url = "http://188.35.161.31/core/deb.php?fdate=" + base_requester; //Генерируем URL для запроса к БД
                let redis_key = "Graphql" + "#" + "Not_calibr" + "#" + "LastMeasurement" + "#" + request_uname + "#" + request_serial; //Генерируем ключ Redis
                return req_not_calibr_lastmeasurement_device(url, redis_key, request_uname, request_serial);
            }
        }
        if (time === 'lastweek') { //Данные за последнюю неделю
            if (uname === undefined && serial === undefined) { //Если uname и serial не определены, то вернуть ВСЕ ИЗМЕРЕНИЯ
                //Формирование времени за последние 7 дней
                let now_week = moment();
                let real_now_week = moment();
                real_now_week = real_now_week.format("YYYY-MM-DD%20HH:mm:ss");
                let past_seven_days = now_week.clone().subtract(7, 'days');
                past_seven_days = past_seven_days.format("YYYY-MM-DD%20HH:mm:ss");
                //Конец
                let base_requester_sevendays = past_seven_days + "&sdate=" + real_now_week;
                let url = "http://188.35.161.31/core/deb.php?fdate=" + base_requester_sevendays; //Генерируем URL для запроса к БД
                let redis_key = "Graphql" + "#" + "Not_calibr#Week"; //Генерируем ключ Redis
                return req_not_calibr_all(url, redis_key);
            }
            if (uname !== undefined && serial !== undefined) { //Если uname и serial прибора были указаны
                let request_uname = uname;
                let request_serial = serial;
                //Формирование времени за последнюю неделю
                let now_week = moment();
                let real_now_week = moment();
                real_now_week = real_now_week.format("YYYY-MM-DD%20HH:mm:ss");
                let past_seven_days = now_week.clone().subtract(7, 'days');
                past_seven_days = past_seven_days.format("YYYY-MM-DD%20HH:mm:ss");
                let base_requester_sevendays = past_seven_days + "&sdate=" + real_now_week;
                //--Конец
                let url = "http://188.35.161.31/core/deb.php?fdate=" + base_requester_sevendays; //Генерируем URL для запроса к БД
                let redis_key = "Graphql" + "#" + "Not_calibr" + "#" + "Week" + "#" + request_uname + "#" + request_serial; //Генерируем ключ Redis
                return req_not_calibr_device(url, redis_key, request_uname, request_serial);
            }
        }
        if (time === 'day') { //Данные за день
            if (uname === undefined && serial === undefined && date !== undefined) { //Если uname и serial не определены, то вернуть ВСЕ ИЗМЕРЕНИЯ
                //Формирование даты
                let day = date;
                let till_next_day = date;
                let first_date = day;
                let second_date = moment(till_next_day).add(24, 'hours').format("YYYY-MM-DD%20HH:mm:ss");
                //Конец
                let base_requester_day = first_date + "&sdate=" + second_date;
                let url = "http://188.35.161.31/core/deb.php?fdate=" + base_requester_day;//Генерируем URL для запроса к базе
                let redis_key = "Graphql" + "#" + "Not_calibr" + "#" + day; //Генерируем ключ Redis
                return req_not_calibr_all(url, redis_key);
            }
            if (uname !== undefined && serial !== undefined && date !== undefined) { //Если uname и serial прибора были указаны
                let request_uname = uname;
                let request_serial = serial;
                //Формирование даты
                let day = date;
                let first_date = day;
                let till_next_day = date;
                let second_date = moment(till_next_day).add(24, 'hours').format("YYYY-MM-DD%20HH:mm:ss");
                //Конец
                let base_requester_day = first_date + "&sdate=" + second_date;
                let url = "http://188.35.161.31/core/deb.php?fdate=" + base_requester_day; //Генерируем URL для запроса к БД
                let redis_key = "Graphql" + "#" + "Not_calibr" + "#" + day + "#" + request_uname + "#" + request_serial; //Генерируем ключ Redis
                return req_not_calibr_device(url, redis_key, request_uname, request_serial);
            }
        }
        if (time === 'log') { //Данные за определенный интервал
            if (uname === undefined && serial === undefined && fdate !== undefined && sdate !== undefined) { //Если uname и serial не определены, то вернуть ВСЕ ИЗМЕРЕНИЯ
                let start_date = fdate; //начальная дата интервала
                let finish_date = sdate;//Конечная дата интервала
                let url = "http://188.35.161.31/core/deb.php?fdate=" + start_date + "&sdate=" + finish_date;//Генерируем URL для бд
                let redis_key = "Graphql" + "#" + "Not_calibr" + "#" + start_date + "#" + finish_date; //Генерируем ключ redis
                return req_not_calibr_all(url, redis_key);
            }
            if (uname !== undefined && serial !== undefined && fdate !== undefined && sdate !== undefined) { //Если uname и serial прибора были указаны
                let request_uname = uname;
                let request_serial = serial;
                let start_date = fdate; //get start_date from request
                let finish_date = sdate;//get finish_date frim request
                let url = "http://188.35.161.31/core/deb.php?fdate=" + start_date + "&sdate=" + finish_date;//Генерируем URL для бд
                let redis_key = "Graphql" + "#" + "Not_calibr" + "#" + start_date + "#" + finish_date + "#" + request_uname + "#" + request_serial;//Генерируем ключ Redis
                return req_not_calibr_device(url, redis_key, request_uname, request_serial);
            }
        }
    },

    calibr: ({time, fdate, sdate, date, uname, serial}) => {
        if (time === 'day') {
            if (uname === undefined && serial === undefined && date !== undefined) { //Если uname и serial не определены, то вернуть ВСЕ ИЗМЕРЕНИЯ
                //Формирование даты
                let day = date;
                let till_next_day = date;
                let first_date = day;
                let second_date = moment(till_next_day).add(24, 'hours').format("YYYY-MM-DD%20HH:mm:ss");
                //Конец
                let base_requester_day = first_date + "&sdate=" + second_date;
                let url = "http://188.35.161.31/core/deb.php?fdate=" + base_requester_day;//Генерация URL для запроса к БД
                let redis_key = "Graphql" + "Calibr" + "#" + day; //Генерация ключа Redis

                return req_calibr_all(url, redis_key);

            }
            if (uname !== undefined && serial !== undefined && date !== undefined) { //Если uname и serial прибора были указаны
                let request_uname = uname;
                let request_serial = serial;
                //Формирование даты
                let day = date;
                let till_next_day = date;
                let first_date = day;
                let second_date = moment(till_next_day).add(24, 'hours').format("YYYY-MM-DD%20HH:mm:ss");
                //Конец
                let base_requester_day = first_date + "&sdate=" + second_date;
                let url = "http://188.35.161.31/core/deb.php?fdate=" + base_requester_day; //Генерация URL для запроса к БД
                let redis_key = "Graphql" + "#" + "Calibr" + "#" + day + "#" + request_uname + "#" + request_serial; //Генерация ключа Redis
                return req_calibr_device(url, redis_key, request_uname, request_serial);
            }
        }
        if (time === 'log') {
            if (uname === undefined && serial === undefined && fdate !== undefined && sdate !== undefined) { //Если uname и serial не определены, то вернуть ВСЕ ИЗМЕРЕНИЯ
                let request_uname = uname;
                let request_serial = serial;
                let start_date = fdate;
                let finish_date = sdate;
                let url = "http://188.35.161.31/core/deb.php?fdate=" + start_date + "&sdate=" + finish_date;//Генерируем URL для бд
                let redis_key = "Graphql" + "#" + "Calibr" + "#" + start_date + "$" + finish_date;//Генерируем ключ Redis
                return req_calibr_all(url, redis_key);

            }
            if (uname !== undefined && serial !== undefined && fdate !== undefined && sdate !== undefined) { //Если uname и serial прибора были указаны
                let request_uname = uname;
                let request_serial = serial;
                let start_date = fdate;
                let finish_date = sdate;
                let url = "http://188.35.161.31/core/deb.php?fdate=" + start_date + "&sdate=" + finish_date;//Генерируем URL для БД
                let redis_key = "Graphql" + "#" + "Calibr" + "#" + start_date + "#" + finish_date + "#" + request_uname + "#" + request_serial;//Генерируем ключ для Redis
                return req_calibr_device(url, redis_key, request_uname, request_serial);
            }
        }

        if (time === 'last5min') {
            if (uname === undefined && serial === undefined) { //Если uname и serial не определены, то вернуть ВСЕ ИЗМЕРЕНИЯ
                //Формирование даты//
                let now = moment();
                let past = now.subtract(5, 'minutes');
                past = past.format("YYYY-MM-DD%20HH:mm:ss");
                let real_now = moment();
                real_now = real_now.format("YYYY-MM-DD%20HH:mm:ss");
                let base_requester = past + "&sdate=" + real_now;
                //---Конец
                let url = "http://188.35.161.31/core/deb.php?fdate=" + base_requester; //Генерируем URL для запроса к БД
                let redis_key = "Graphql" + "#" + "Calibr" + "#" + "LastLog"; //Генерируем ключ Redis
                return req_calibr_all(url, redis_key);

            }
            if (uname !== undefined && serial !== undefined) { //Если uname и serial прибора были указаны
                let request_uname = uname;
                let request_serial = serial;
                //Формирование даты
                let now = moment();
                let past = now.subtract(5, 'minutes');
                past = past.format("YYYY-MM-DD%20HH:mm:ss");
                let real_now = moment();
                real_now = real_now.format("YYYY-MM-DD%20HH:mm:ss");
                //---Конец
                let base_requester = past + "&sdate=" + real_now;
                let url = "http://188.35.161.31/core/deb.php?fdate=" + base_requester; //Генерируем URL для запроса к БД
                let redis_key = "Graphql" + "#" + "Calibr" + "#" + "LastLog" + "#" + request_uname + "#" + request_serial; //Генерируем ключ Redis
                return req_calibr_device(url, redis_key, request_uname, request_serial);
            }
        }
        if (time === 'lastmin') {
            if (uname === undefined && serial === undefined) { //Если uname и serial не определены, то вернуть ВСЕ ИЗМЕРЕНИЯ
                //Формирование даты//
                let now = moment();
                let past = now.subtract(1, 'minutes');
                past = past.format("YYYY-MM-DD%20HH:mm:ss");
                let real_now = moment();
                real_now = real_now.format("YYYY-MM-DD%20HH:mm:ss");
                let base_requester = past + "&sdate=" + real_now;
                //---Конец
                let url = "http://188.35.161.31/core/deb.php?fdate=" + base_requester; //Генерируем URL для запроса к БД
                let redis_key = "Graphql" + "#" + "Calibr" + "#" + "LastMin"; //Генерируем ключ Redis
                return req_calibr_all(url, redis_key);

            }
            if (uname !== undefined && serial !== undefined) { //Если uname и serial прибора были указаны
                let request_uname = uname;
                let request_serial = serial;
                //Формирование даты
                let now = moment();
                let past = now.subtract(1, 'minutes');
                past = past.format("YYYY-MM-DD%20HH:mm:ss");
                let real_now = moment();
                real_now = real_now.format("YYYY-MM-DD%20HH:mm:ss");
                //---Конец
                let base_requester = past + "&sdate=" + real_now;
                let url = "http://188.35.161.31/core/deb.php?fdate=" + base_requester; //Генерируем URL для запроса к БД
                let redis_key = "Graphql" + "#" + "Calibr" + "#" + "LastMin" + "#" + request_uname + "#" + request_serial; //Генерируем ключ Redis
                return req_calibr_device(url, redis_key, request_uname, request_serial);
            }
        }
        if (time === 'last_measurement') {
            if (uname === undefined && serial === undefined) { //Если uname и serial не определены, то вернуть ВСЕ ИЗМЕРЕНИЯ
                //Формирование даты//
                let now = moment();
                let past = now.subtract(60, 'minutes');
                past = past.format("YYYY-MM-DD%20HH:mm:ss");
                let real_now = moment();
                real_now = real_now.format("YYYY-MM-DD%20HH:mm:ss");
                let base_requester = past + "&sdate=" + real_now;
                //---Конец
                let url = "http://188.35.161.31/core/deb.php?fdate=" + base_requester; //Генерируем URL для запроса к БД
                let redis_key = "Graphql" + "#" + "Calibr" + "#" + "LastMeasurement"; //Генерируем ключ Redis
                return req_calibr_lastmeasurement_all(url, redis_key);
            }
            if (uname !== undefined && serial !== undefined) { //Если uname и serial прибора были указаны
                let request_uname = uname;
                let request_serial = serial;
                //Формирование даты
                let now = moment();
                let past = now.subtract(60, 'minutes');
                past = past.format("YYYY-MM-DD%20HH:mm:ss");
                let real_now = moment();
                real_now = real_now.format("YYYY-MM-DD%20HH:mm:ss");
                //---Конец
                let base_requester = past + "&sdate=" + real_now;
                let url = "http://188.35.161.31/core/deb.php?fdate=" + base_requester; //Генерируем URL для запроса к БД
                let redis_key = "Graphql" + "#" + "Calibr" + "#" + "LastMeasurement" + "#" + request_uname + "#" + request_serial; //Генерируем ключ Redis
                return req_calibr_lastmeasurement_device(url, redis_key, request_uname, request_serial);
            }
        }
        if (time === 'lastweek') {
            if (uname === undefined && serial === undefined) { //Если uname и serial не определены, то вернуть ВСЕ ИЗМЕРЕНИЯ
                //Генерация даты за последнюю неделю
                let now_week = moment();
                let real_now_week = moment();
                real_now_week = real_now_week.format("YYYY-MM-DD%20HH:mm:ss");
                let past_seven_days = now_week.clone().subtract(7, 'days');
                past_seven_days = past_seven_days.format("YYYY-MM-DD%20HH:mm:ss");
                let base_requester_sevendays = past_seven_days + "&sdate=" + real_now_week;
                //---Конец
                let url = "http://188.35.161.31/core/deb.php?fdate=" + base_requester_sevendays; //Генерация url для запроса к БД
                let redis_key = "Graphql" + "#" + "Calibr" + "#" + "Week"; //Генерация ключа Redi
                return req_calibr_all(url, redis_key);

            }
            if (uname !== undefined && serial !== undefined) { //Если uname и serial прибора были указаны
                let request_uname = uname;
                let request_serial = serial;
                //Генерация даты за последнюю неделю
                let now_week = moment();
                let real_now_week = moment();
                real_now_week = real_now_week.format("YYYY-MM-DD%20HH:mm:ss");
                let past_seven_days = now_week.clone().subtract(7, 'days');
                past_seven_days = past_seven_days.format("YYYY-MM-DD%20HH:mm:ss");
                let base_requester_sevendays = past_seven_days + "&sdate=" + real_now_week;
                //---Конец
                let url = "http://188.35.161.31/core/deb.php?fdate=" + base_requester_sevendays; //Генерация URL для запроса к БД
                let redis_key = "Graphql" + "#" + "Calibr" + "#" + "Week" + "#" + request_uname + "#" + request_serial; //Генерация ключа Redis
                return req_calibr_device(url, redis_key, request_uname, request_serial);
            }
        }

    }

};
var app = express();
//Указываем описанную выше схему, для удобства отладки делаем graphiql = true
const apiRouter = require('./routes/Rest');//connect route/api
app.use('/api', apiRouter); //use api route
app.use('/graphql', express_graphql({
    schema: schema,
    rootValue: root,
    graphiql: true
}));
app.listen(4000, () => console.log('Express GraphQL Server Now Running On localhost:4000/graphql')); //Запускаем программный интерфейс, используя порт 4000

module.exports = app;


